#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Set encoding default for python 2.7
# vim: syntax=python noexpandtab

# gadgetconfig configures and enables a Gadget USB configuration
#

import os
import fnmatch

"""gadget.py: ..."""

# __author__  = "Stuart.Lynne@belcarra.com"


class AddGadget(object):

	def __init__(self, configpath, verbose=False):
		self.configpath = configpath
		self.verbose = verbose
		# self.verbose = True

	def vprint(self, s):
		if self.verbose:
			print(s)

	def write_str(self, path, s):
		self.vprint("write_str: %s \"%s\"" % (path, s.strip()))
		try:
			f = open(path, "a")
			f.writelines(s)
			f.close()
		except (PermissionError):
			print("%s %s PERMISSION DENIED" % (path, s.strip()))

	def makedirs(self, lpath):
		self.vprint("makedirs: %s" % (lpath))
		os.makedirs(lpath)

	def symlink(self, src, target):
		self.vprint("symlink: %s -> %s" % (target, src))
		os.symlink(src, target)

	def hex_or_str(self, v):
		if isinstance(v, (int)):
			return "0x%02x\n" % v
		else:
			return "%s\n" % v

	# create_strings
	# For each language create a stings/lang directory containing
	# a file for each string descriptor. E.g.
	#
	#   strings/0x409/
	#   strings/0x409/manfacturer
	#   strings/0x409/product
	#
	def create_strings(self, path, string_dicts):
		exclude = ['#*']
		for lang in string_dicts:
			if any(fnmatch.fnmatch(lang, pattern) for pattern in exclude):
				continue
			spath = "%s/strings/%s" % (path, lang)
			self.makedirs(spath)
			string_dict = string_dicts[lang]
			for s in string_dict:
				self.write_str("%s/%s" % (spath, s), "%s\n" % string_dict[s])

	# add_attrs
	# Add misc attributes to a path, convert ints to hex strings,
	# ignoring dictionary entries with values that are not int or str.
	#
	def add_attrs(self, path, dict, exclude=[]):
		print("add_attrs: path: %s dict: %s" % (path, dict))
		exclude.append("#*")
		for a in dict:
			if any(fnmatch.fnmatch(a, pattern) for pattern in exclude):
				continue
			if isinstance(dict[a], (int, str)):
				self.write_str("%s/%s" % (path, a), self.hex_or_str(dict[a]))

	# create_device_os_desc
	# Create the Gadget Device os_descs directory, with symlink to
	# configuration.
	def create_device_os_desc(self, path, device_definition):

		try:
			os_descs_dict = device_definition['os_descs']
		except (KeyError):
			return

		# print("create_device_os_descs_dict: %s" % (os_descs_dict))
		if os_descs_dict is None:
			return

		lpath = "%s/os_desc" % (path)
		self.makedirs(lpath)

		# See if we have separate config_id and config_name, use
		# to compose config name to build src and target paths.
		try:
			config_id = os_descs_dict['config_id']
			config_name = os_descs_dict['config_name']
			config = "%s.%s" % (config_name, config_id)
			src = "/%s/configs/%s" % (path, config)
			target = "%s/%s" % (lpath, config)
			self.symlink(src, target)
		except KeyError:
			pass

		self.add_attrs(lpath, os_descs_dict, exclude=['config_id', 'config_name'])

	# create_functions
	# Create the Gadget Device Functions
	def create_functions(self, path, functions_dict):

		# print("create_functions: %s" % (functions_dict))

		functions_path = "%s/functions" % (path)
		exclude = ['#*']
		for function_name in functions_dict:
			if any(fnmatch.fnmatch(function_name, pattern) for pattern in exclude):
				continue

			# print("create_functions: %s %s" % (function_name, functions_dict[function_name]))

			function_dict = functions_dict[function_name]
			function_path = "%s/%s" % (functions_path, function_name)
			self.makedirs(function_path)
			self.add_attrs(function_path, function_dict)

			# os_descs is optional, may not be present
			if 'os_descs' not in function_dict:
				continue
			function_os_descs = function_dict['os_descs']
			for interface in function_os_descs:
				ipath = "%s/os_desc/%s" % (function_path, interface)
				print("create_functions: ipath: %s" % (ipath))

	# create_configs
	# Create the Gadget Device Configurations
	def create_configs(self, path, configs_dict):

		print("create_configs: %s" % (configs_dict))

		configs_path = "%s/configs" % (path)

		for config_name in configs_dict:
			config_dict = configs_dict[config_name]
			print("create_configs: %s" % (config_name))
			print("create_configs: %s" % (config_dict))

			# create directory and add the attributes and strings
			config_path = "%s/%s" % (configs_path, config_name)
			self.makedirs(config_path)
			self.add_attrs(config_path, config_dict)
			self.create_strings(config_path, config_dict['strings'])

			# add the function symlinks
			function_dict = config_dict['functions']
			for f in function_dict:
				print("create_configs: f: %s" % (f))
				function = f['function']
				target = "%s/%s" % (config_path, f['name'])
				src = "%s/functions/%s" % (path, function.replace("_", "."))
				# target = "%s/%s" % (config_path, f)
				self.symlink(src, target)

	# add_device_json
	# Create a Gadget Device definition from saved json configuratiion
	#
	def add_device_json(self, device_definition, device_name=None):
		device_path = "%s/%s" % (self.configpath, device_name)
		self.makedirs(device_path)

		# handle device attributes, anything at the top level that is a string
		# or int, not a dict
		self.add_attrs(device_path, device_definition)

		# add the strings, functions, configs and os_desc
		#
		self.create_strings(device_path, device_definition['strings'])
		self.create_functions(device_path, device_definition['functions'])
		self.create_configs(device_path, device_definition['configs'])
		self.create_device_os_desc(device_path, device_definition)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Set encoding default for python 2.7
# vim: syntax=python noexpandtab

# gadgetconfig configures and enables a Gadget USB configuration
#

import os
import collections
from datetime import date
import fnmatch

"""gadget.py: ..."""

# __author__  = "Stuart.Lynne@belcarra.com"


class ExportGadget(object):

	def __init__(self, configpath, verbose=False):
		self.configpath = configpath
		self.verbose = verbose
		self.verbose = True

	def pathread(self, path):

		# print("pathread: %s" % (path))
		try:
			fstat = os.stat(path)
			# print("fstat: size:%s" % (fstat.st_size))
		except (PermissionError):
			print("pathread: %s PERMISSION ERROR" % (path))
			return ''
		except (FileNotFoundError):
			print("pathread: %s NOT FOUND" % (path))
			return ''

		# 4096 or 0 byte files should contain info
		if fstat.st_size == 4096 or fstat.st_size == 0:
			try:
				f = open(path, "r")
				lines = f.readlines(1000)
				f.close()
				return lines
			except (PermissionError, OSError):
				return ''
			except UnicodeDecodeError:
				return '[UnicodeDecodeError]'

		return '<UNKNOWN>'

	def export_attribute_entry(self, path, entry, symlinks=False, exclude=[]):
		if any(fnmatch.fnmatch(entry, pattern) for pattern in exclude):
			return None
		# print("export_attributes: entry %s" % (entry))
		epath = "%s/%s" % (path, entry)
		if symlinks and os.path.islink(epath):
			realpath = os.path.realpath(epath)
			(head, tail) = os.path.split(realpath)
			return tail
		if os.path.isdir(epath) or os.path.islink(epath) or not os.path.isfile(epath):
			return None
		return self.pathread(epath)[0].rstrip('\t\r\n\0')

	def export_attribute(self, path, attributes, entry):
		if entry not in attributes:
			return None
		return self.export_attribute_entry(path, entry)

	def export_attributes(self, path, symlinks=False, exclude=[], idFlag=False, annotation=None):

		attributes = [{}, annotation][annotation is not None]

		# print("export_attributes: annotation %s" % (annotation), file=sys.stderr)
		# print("export_attributes: attributes %s" % (attributes), file=sys.stderr)
		attribute_entries = sorted(os.listdir(path), key=str.casefold)

		idList = ['idVendor', 'idProduct', 'bcdDevice', 'bDeviceClass', 'bDeviceSubClass', 'bDeviceProtocol']
		# print("export_attributes: entries %s" % (attribute_entries), file=sys.stderr)

		if idFlag:
			for entry in idList:
				v = self.export_attribute(path, attribute_entries, entry)
				if v is not None:
					attributes[entry] = v

		for entry in attribute_entries:
			if idFlag and entry in idList:
				continue
			v = self.export_attribute_entry(path, entry, symlinks, exclude)
			if v is not None:
				attributes[entry] = v
		# print("export_attributes: %s" % (attributes), file=sys.stderr)
		# print("", file=sys.stderr)
		return attributes

	def export_strings(self, strings_path, annotation=None):
		strings = [{}, annotation][annotation is not None]
		# print("export_strings: strings_path %s" % (strings_path))
		for lang_name in sorted(os.listdir(strings_path), key=str.casefold):
			# print("export_strings: lang_name %s" % (lang_name))
			string_path = "%s/%s" % (strings_path, lang_name)
			# print("export_strings: string_path %s" % (string_path))
			strings[lang_name] = self.export_attributes(string_path)
		return strings

	def export_device_configs(self, configs_path):
		print("***********************************")
		print("export_device_configs: configs_path: %s" % (configs_path))
		configs = {}
		for config_name in sorted(os.listdir(configs_path), key=str.casefold):
			functions = []
			print("-------")
			print("export_device_configs: config_name: %s" % (config_name))
			config_path = "%s/%s" % (configs_path, config_name)
			config_entries = sorted(os.listdir(config_path), key=str.casefold)
			config = self.export_attributes(config_path, symlinks=True,
					annotation={
						'# Configuration Descriptor': '',
						'# bmAttributes: bit 5 support remote wakeup': '',
						'# bmAttributes: bit 6 self-powered': '',
						'# bmAttributes: bit 7 bus-powered': '',
						'# MaxPower: Power requirements in two-milliampere units, only valid of bit 7 is set': '',
					})
			for entry in config_entries:
				epath = "%s/%s" % (config_path, entry)
				print("export_device_configs: epath: %s" % (epath))
				if os.path.isdir(epath) and not os.path.islink(epath):
					# is a directory
					if entry == 'strings':
						config[entry] = self.export_strings(epath,
								annotation={'# USB Device Configuration Strings': ''})
					continue
				elif os.path.islink(epath):
					# symlink - should not be any at this level
					realpath = os.path.realpath(epath)
					(path, target) = os.path.split(realpath)
					functions.append({'name': entry, 'function': target})
					continue
				elif os.path.isfile(epath):
					# should be regular file
					continue
				else:
					# unknown!
					continue

			config['functions'] = functions
			configs[config_name] = config

		return configs

	def export_function_os_desc(self, os_desc_path):
		# print("")
		# print("export_function_os_desc: os_desc_path: %s" % (os_desc_path))
		os_desc = {}
		os_desc_entries = sorted(os.listdir(os_desc_path), key=str.casefold)
		for entry in os_desc_entries:
			epath = "%s/%s" % (os_desc_path, entry)
			if os.path.isdir(epath) and not os.path.islink(epath) and fnmatch.fnmatch(entry, "interface.*"):
				os_desc[entry] = self.export_attributes(epath)
		return os_desc

	def export_device_functions(self, functions_path, annotation=None):
		# print("")
		# print("export_device_functions: functions_path: %s" % (functions_path))
		functions = [{}, annotation][annotation is not None]
		for function_name in sorted(os.listdir(functions_path), key=str.casefold):
			function_path = "%s/%s" % (functions_path, function_name)
			function_entries = sorted(os.listdir(function_path), key=str.casefold)
			function = self.export_attributes(function_path, exclude=['ifname', 'port_num'])
			for entry in function_entries:
				epath = "%s/%s" % (function_path, entry)
				if os.path.isdir(epath) and not os.path.islink(epath):
					# is a directory
					if entry == 'os_desc':
						function[entry] = self.export_function_os_desc(epath)
					continue
				elif os.path.islink(epath):
					# sysmlink - should not be any at this level
					realpath = os.path.realpath(epath)
					target = os.path.split(realpath)
					function[entry] = target
					continue
				elif os.path.isfile(epath):
					# should be regular file
					continue
				else:
					# unknown!
					continue

			functions[function_name] = function
			pass
		return functions

	def export_device_os_desc(self, os_desc_path):
		# print("export_device_os_desc: os_desc_path: %s" % (os_desc_path))
		os_desc = self.export_attributes(os_desc_path, symlinks=True)
		return os_desc

	def export_devices(self):
		device_paths = sorted(os.listdir(self.configpath), key=str.casefold)
		devices = collections.OrderedDict()
		devices['# Gadget Device Definition File'] = ''
		devices["# %s" % (date.today())] = ''
		for device_name in device_paths:
			device_path = "%s/%s" % (self.configpath, device_name)
			device = self.export_attributes(device_path, exclude=['UDC'], idFlag=True,
					annotation={'# USB Device Descriptor Fields': ''})
			# print("export_device: device_path: %s" % (device_path))
			device_entries = sorted(os.listdir(device_path), key=str.casefold)
			if 'strings' in device_entries:
				epath = "%s/%s" % (device_path, 'strings')
				device['# USB Device Strings'] = ''
				device['strings'] = self.export_strings(epath)
			if 'functions' in device_entries:
				epath = "%s/%s" % (device_path, 'functions')
				device['# Gadget Functions list: see /sys/module/usb_f*,'] = ''
				device['# E.g.: usb_f_acm, usb_f_ecm, usb_f_eem, usb_f_hid, usb_f_mass_storage'] = ''
				device['#       usb_f_midi, usb_f_ncm, usb_f_obex, usb_f_rndis, usb_f_serial'] = ''
				device['# Use: The function name (without prefix) is used with instantion name, e.g. eem.usb0 or acm.GS0'] = ''
				device['functions'] = self.export_device_functions(epath)
			if 'configs' in device_entries:
				epath = "%s/%s" % (device_path, 'configs')
				device['# Gadget Configurations list'] = ''
				device['configs'] = self.export_device_configs(epath)
			for entry in device_entries:
				# print("export_device: device_path: %s entry: %s" % (device_path, entry))
				epath = "%s/%s" % (device_path, entry)
				if os.path.isdir(epath) and not os.path.islink(epath):
					# is a directory
					if entry == 'configs':
						pass
					elif entry == 'functions':
						pass
					elif entry == 'os_desc':
						device['# Microsoft OS Descriptors Support'] = ''
						device['# C.f. https://docs.microsoft.com/en-us/previous-versions/gg463179(v=msdn.10)'] = ''
						device[entry] = self.export_device_os_desc(epath)
					elif entry == 'strings':
						pass
				elif os.path.islink(epath):
					# sysmlink - should not be any at this level
					pass
				# elif os.path.isfile(epath):
					# should be regular file
					pass
				else:
					# unknown!
					pass

			devices[device_name] = device

		return devices




from setuptools import setup


def main():

    setup(
        name='gadgetconfig',
        packages=['gadgetconfig'],
        package_dir={'': 'src'},
        version=open('VERSION.txt').read().strip(),
        author='Stuart Lynne',
        author_email='stuart.lynne@belcarra.com',
        url='http://github.com/Belarra/gadgetconfig',
        download_url='http://github.com/Belarra/gadgetconfig.git',
        license='MIT',
        keywords=['usb', 'gadget'],
        description='gadgetconfig creates and controls Gadget USB Devices and integrates Gadget with systemd',
        entry_points={'console_scripts': ['gadgetconfig = gadgetconfig:main', ], },
        py_modules=['add', 'export', 'manage', 'remove'],
        install_requires=["argparse"],
        classifiers=[
            "Programming Language :: Python",
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: System Administrators",
            "Operating System :: POSIX",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            'Topic :: System :: Logging',
            'Topic :: Text Processing',
            "Topic :: Software Development :: Libraries :: Python Modules",
            "Topic :: System :: System Shells",
            "Topic :: System :: Systems Administration",
        ],
        data_files=[
            ('/etc/gadgetservice', ['definitions/belcarra-2acm+ecm.json']),
            ('/usr/lib/gadgetservice', ['service/gadget.start', 'service/gadget.stop']),
            ('/lib/systemd/system', ['service/gadget.service']),
            ('/usr/share/doc/gadgetconfig', ['README.md', 'README-Gadget.md', 'README-Raspbian.md']), ],

        long_description=open('README.md').read(),
        long_description_content_type='text/markdown'

    )


if __name__ == '__main__':
    main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Set encoding default for python 2.7
# vim: syntax=python noexpandtab

# gadgetconfig configures and enables a Gadget USB configuration
#

import os
import sys

import io
import re
import argparse
import json
import commentjson

try:
	from gadgetconfig.add import AddGadget
	from gadgetconfig.export import ExportGadget
	from gadgetconfig.manage import ManageGadget
	from gadgetconfig.remove import RemoveGadget
except ModuleNotFoundError:
	from add import AddGadget
	from export import ExportGadget
	from manage import ManageGadget
	from remove import RemoveGadget


"""gadgetconfig.py: ..."""

# __author__  = "Stuart.Lynne@belcarra.com"


def remove_old_device(self, configname):

	fpath = "%s/%s" % (self.configpath, configname)

	print("Check for old configuration: %s" % (fpath))

	if not os.path.exists(fpath) and not os.path.lexists(fpath):
		print("%s does not exist" % (fpath))
		return

	if not os.path.isdir(fpath):
		print("ERROR %s exists and is not a directory" % (fpath))
		exit(1)

	exit(1)


def replace(data, match, repl):
	if isinstance(data, (dict, list)):
		for k, v in (data.items() if isinstance(data, dict) else enumerate(data)):
			if k == match:
				data[k] = repl
			replace(v, match, repl)


# this is mainly for testing standalone
#
def main():
	parser = argparse.ArgumentParser(
		description="Configure Gadget Device using SysFS and ConfigFS",
		formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=999))

	parser.add_argument("-T", "--test", help="run tests", action='store_true')
	# parser.add_argument("-L", "--conf", nargs='?', help="include (shell pattern match)", default='')
	# parser.add_argument("paths", metavar='Path', type=str, nargs=argparse.REMAINDER, help="pathname", default=[])

	# parser.add_argument("-D", "--disable", help="Disable the current Gadget", action='store_true')
	# parser.add_argument("path", metavar='path', nargs='?', type=str, help="pathname", default=None)
	# parser.add_argument("-A", "--attr", help="Dump Scheme Attr", action='store_true')
	# parser.add_argument("-2", "--scheme2json", help="Scheme Attr to JSON", action='store_true')
	# parser.add_argument("-S", "--scheme", help="Scheme Attr to JSON", action='store_true')

	parser.add_argument("--idVendor", type=str, help="Optional device idVendor attribute")
	parser.add_argument("--idProduct", type=str, help="Optional device idProduct attribute")

	parser.add_argument("--manufacturer", type=str, help="Optional device manufacturer string")
	parser.add_argument("--product", type=str, help="Optional device product string")
	parser.add_argument("--serialnumber", type=str, help="Optional device serialnumber string")

	parser.add_argument("--dev_addr", type=str, help="Optional ecm dev_addr attribute")
	parser.add_argument("--host_addr", type=str, help="Optional ecm host_addr attribute")

	parser.add_argument("-N", "--name", nargs='?', type=str, help='device name override', default=None)
	# parser.add_argument("-I", "--id", nargs='?', type=int, help='enable ID', default=1)
	parser.add_argument("-X", "--export", help="Export JSON to STDOUT", action='store_true')

	group = parser.add_mutually_exclusive_group(required=False)

	parser.add_argument("--soft-disconnect", "--soft_disconnect", help="Disable the current Gadget", action='store_true')
	parser.add_argument("--soft-connect", "--soft_connect", help="Disable the current Gadget", action='store_true')

	group.add_argument("-A", "--add", type=str, help="Add Gadget device definitions from JSON file", default=None)
	group.add_argument("-R", "--remove", nargs='?', type=str, help='Remove Gadget device definition', default=None)

	group.add_argument("-E", "--enable", type=str, help="enable specified Gadget", default=None)
	group.add_argument("-D", "--disable", help="disable current Gadget", action='store_true')
	# group.add_argument("path", metavar='path', nargs='?', type=str, help="Gadget Device Scheme", default=None)

	args = parser.parse_args()

	print("args: %s" % (args), file=sys.stderr)
	print("", file=sys.stderr)

	if args.test:
		sys_config_path = "sys/kernel/config/usb_gadget"
	else:
		sys_config_path = "/sys/kernel/config/usb_gadget"

	m = ManageGadget(sys_config_path)
	if not m.checkfs(verbose=False):
		if not args.test:
			exit(1)
		os.makedirs("sys/kernel/config/usb_gadget")

	m.find_udcs(verbose=True)
	m.check_current(verbose=False)

	if args.disable:
		m.disable_current()
		exit(0)

	if args.enable is not None:
		m.enable_current(args.enable)
		exit(0)

	if args.soft_disconnect:
		m.soft_connect(False)
		exit(0)

	if args.soft_connect:
		m.soft_connect(True)
		exit(0)

	if args.export:
		e = ExportGadget(sys_config_path)
		devices = e.export_devices()
		j = json.dumps(devices, indent=4)
		print(re.sub(r'"(#.*)":.*,', r'\1', j))
		exit(0)

	if args.remove is not None:
		r = RemoveGadget(sys_config_path)
		r.remove_device(args.remove)
		exit(0)

	if args.add is not None:
		a = AddGadget(sys_config_path)
		with io.open(args.add) as f:
			device_definitions = commentjson.load(f)
			for device_name in device_definitions:
				device_definition = device_definitions[device_name]
				if args.idVendor: replace(device_definition, 'idVendor', args.idVendor)
				if args.idProduct: replace(device_definition, 'idProduct', args.idProduct)
				if args.manufacturer: replace(device_definition, 'manufacturer', args.manufacturer)
				if args.product: replace(device_definition, 'product', args.product)
				if args.serialnumber: replace(device_definition, 'serialnumber', args.serialnumber)
				if args.dev_addr: replace(device_definition, 'dev_addr', args.dev_addr)
				if args.host_addr: replace(device_definition, 'host_addr', args.host_addr)
				print('device_name: %s' % (device_name))
				if args.name is not None:
					device_name = args.name
					print('device_name: %s' % (device_name))
				a.add_device_json(device_definition, device_name=device_name)

	exit(1)


if __name__ == "__main__":
	main()
